/*======================================================*/
/*==================== audio.c ==========================*/
/*=======================================================*/

#define STANDALONE 1 //enable all necessary from hardware.h

#include <stdio.h>
#include <hardware.h>
#include <string.h>
#include <stdlib.h>
#include <vstypes.h>
#include <parametric.h>

//#include <math.h>

#define BLOCKSIZE 16 //Number of stereo samples processed in one loop round

void Disable(void);	/* NOT auto */
void Enable(void);	/* NOT auto */

extern s_int16 (*applAddr)(s_int16 register __i0 **d,
			   s_int16 register __a1 mode,
			   s_int16 register __a0 n);

auto s_int16 WmaStereoCopy(s_int16 *buf, s_int16 cnt);
auto void MyGetCPairs(register __i0 __y u_int16 *p, register __a0 s_int16 n);


// put u_int16 as hex to serial console for debugging
__y const char hex[] = "0123456789abcdef";
// Note: VS1000A can't have inited y-vars in NAND boot
void puthex(u_int16 a) {
  char tmp[8];
  tmp[0] = hex[(a>>12)&15];
  tmp[1] = hex[(a>>8)&15];
  tmp[2] = hex[(a>>4)&15];
  tmp[3] = hex[(a>>0)&15];
  tmp[4] = ' ';
  tmp[5] = '\0';
  fputs(tmp, stdout);
}

/*
  Called before HALT, i.e. when there is nothing else to do.
  Either there is not enough data in stream buffer, or 
  the audio buffer is full.
 */
void UserHook(void) {
}

#define DELAY_BUF_SZ 8192

s_int16 __y buffer[DELAY_BUF_SZ]; //delay line

void main(void) {
  s_int16 audioSamp[2*BLOCKSIZE];
  s_int16 __y lineInBuf[2*BLOCKSIZE];

  s_int16 __y *bufptr = buffer;
  s_int16 __y *flgptr = buffer;

 
  s_int16 phase = 0;
  s_int16 flangerSample = 0;
  u_int16 flangerPhase = 0;

  /*
  max delay = 8192/24000 = 341ms

  flanger = 5ms .. 20ms (= ~100 .. ~500)
  chorus = 15ms .. 100ms ( = ~500 .. ~2500
  delay = 100ms .. 341ms ( = ~2500 .. 8195)

  depth = depth of modulation (effect to delay length)
  delay = pre-delay (depth adds to this delay)
  speed = speed of modulation
    - add speed to phase every 64 (BLOCKSIZE) samples (24000Hz samplerate) -> 375 per/sec 
    - minimum frequency is 1/(2^16/375) = 0.00572204Hz
    - step is 0.00572204Hz
    - range 0 .. 1024 (5.7Hz)
  decay = amount of feedback from output to input ("repeat" control for delay effect)
  mix   = mix of dry and effect signal (0 == all dry, 32000 == all effect)

  flanger:
   depth = 1 .. 300
   speed = 1 .. 50
   delay = 0
   dacay = 0 .. 32000
   mix = 0 .. 32000

  delay:
   depth = 0
   speed =0
   delay = 300 .. 8192
   dacay = 0 .. 32000
   mix = 0 .. 32000

  "chorus":
   depth = 100 .. 1024
   speed 1 .. 50
   delay = 200 .. ~1000
   dacay = 0 .. 32000
   mix = 0 .. 32000


   22 words of free space in parametric struct
   parametric_x.i.generic[0..3] //reserved
   parametric_x.i.generic[4] == depth
   parametric_x.i.generic[5] == speed
   parametric_x.i.generic[6] == delay
   parametric_x.i.generic[7] == decay
   parametric_x.i.generic[8] == mix
   

  */

#if 1
  //delay
  u_int16 depth = 0; //0 .. 2048
  u_int16 speed = 0; //1 .. 100
  u_int16 delay = 2000; //0 .. 8192 
  u_int16 decay = 16000; //0 .. 32000
  u_int16 mix = 16000;  //0 .. 32000

#else
  //flanger/chorus
  u_int16 depth = 10; //0 .. 2048
  u_int16 speed = 10; //1 .. 100 //4294967296
  u_int16 delay = 0; //0 .. 8192 
  u_int16 decay = 20000; //0 .. 32000
  u_int16 mix = 16000;  //0 .. 32000
#endif

  delay += (depth/2)+1;

  //puts("mo");

  //clear the delay line
  memsetY(buffer, 0, DELAY_BUF_SZ);

  Disable(); //Don't allow interrupts yet

    /* zero applAddr so we are not called again */
    applAddr = 0;

    /* must disable analog powerdown if it isn't already! */
    USEX(SCI_STATUS) &= ~((1<<SCIST_APDOWN1) | (1<<SCIST_APDOWN2));
    
    /*==========================================*/
    /*= Turn left analog input to line-in mode =*/
    /*==========================================*/
#if 0
    /* LINE1 */
    USEX(SCI_MODE) |= (1<<SCIMB_LINE); /*left=line-in, right=line-in*/
#else
    /* MIC */
    USEX(SCI_MODE) &= ~(1<<SCIMB_LINE); /*left=mic, right=line-in*/
#endif

    /*================================================*/
    /*= Set uart speed for serial port communication =*/
    /*================================================*/

    //uartByteSpeed = 11520U;
    
    /*======================*/
    /*= Set DSP core clock =*/
    /*======================*/
    //USEX(SCI_CLOCKF) = 0x6000; /* 3.0x 12.288MHz */
    //USEX(SCI_CLOCKF) = 0x8000; /* 3.5x 12.288MHz */
    //USEX(SCI_CLOCKF) = 0xa000; /* 4.0x 12.288MHz */
    USEX(SCI_CLOCKF) = 0xc000; /* 4.5x 12.288MHz */
    //USEX(SCI_CLOCKF) = 0xe000; /* 5.0x 12.288MHz */
    SetClockFromSpi();

    USEX(SCI_AICTRL1) = 1024; /*autogain==0, 1x==1024, 0.5x==512 etc.*/
    USEX(SCI_AICTRL3) = 0x0004; /*Joint stereo (common gain). Linear pcm*/

    //adjust AD sample rate

    //USEX(DECIM_CONTROL) = DECIM_ENABLE | DECIM_FACTOR48K; /*AD sample rate is 48 kHz*/
    USEX(DECIM_CONTROL) = DECIM_ENABLE | DECIM_FACTOR24K; /*AD sample rate is 48 kHz*/
    
    //SetHardware(2, 48000U); /*DA sample rate in Hz*/
    SetHardware(2, 24000U); /*DA sample rate in Hz*/

    /* 'Empty' stream buffer*/
    stream_rd_pointer = stream_wr_pointer = (u_int16 __x *)&stream_buffer[0];
    USEX(INT_ENABLE) |= (1<<INT_EN_MODU)/*AD*/ | (1<<INT_EN_DAC) | (1<<INT_EN_RX);

    //max volume
    USEX(SCI_VOL) = 0x0101;

    /* Set audio_buffer read and write pointers */
    audio_rd_pointer = audio_buffer;
    audio_wr_pointer = audio_buffer + 2*BLOCKSIZE;

    //Write zero to audio buffer. Prevent unwanted samples to be played.
    memsetY(audio_buffer, 0, AUDIO_BUFFER_SZ);

    Enable(); //allow interrupts (DAC, AD,...)

    while (1) { //Main loop
#if 0
      USEX(SCI_MODE) |= (1<<SCIMB_LINE); /*left=line-in, right=line-in*/
#endif
      if(StreamDiff() > BLOCKSIZE*2) {
	/*Read input samples*/
	MyGetCPairs((__y u_int16 *)lineInBuf, 2*BLOCKSIZE); //Stereo samples!!!!

	{
	  register u_int16 i;
	  s_int16 temp = 0;
	 
	  u_int32 fractDelay;
	  register __y s_int16 *lp = lineInBuf;
	  register s_int16 *sp = audioSamp;

	  //64 samples at one go
	  for(i=0;i<BLOCKSIZE;i++) { //
	    s_int16 flangerOut = ((s_int32)*flgptr * (65535U-flangerPhase)) >> 16;

	    //wrap buffer and advance flgptr 
	    if (flgptr >= &buffer[sizeof(buffer)-1]) {
	      flgptr = buffer;
	    } else {
	      flgptr++;
	    }

	    flangerOut += ((s_int32)*flgptr * flangerPhase) >> 16;

	    //add input sample to buffer w/ feedback (decay)
	    //mix dry and effect and send to output (mix)
#if 1
	    //lp[0] is left channel (mic or line), lp[1] is right channel
	    *sp = ((s_int32)lp[0] * (32000U-mix) + (s_int32)flangerOut * mix) >> 16 ;
	    //*bufptr = ((s_int32)*sp++ * (32000-decay)) >> 15;
	    *bufptr = *sp++;
#else
	    // this branch does not work yet..
	    *sp = ((s_int32)lp[0] * (32000U-mix) + (s_int32)flangerOut * mix) >> 15;
	    *bufptr = ((s_int32)*sp++ * decay + (s_int32)*lp++ * (32000U-decay)) >> 15;
#endif
	    lp += 2; //samples only in one channel -> offset
	    *sp++ = sp[-1]; // to other output also

	    //temp = *flgptr/2;

	    //wrap buffer and advance bufptr
	    if (bufptr  >= &buffer[sizeof(buffer)-1]) {
	      bufptr = buffer;
	    } else {
	      bufptr++;
	    }

	    //temp += *flgptr/2;
	    
	  }
	  
	  //LFO (speed & depth)
	  phase += speed; 

	  // LFO value varies from 0 to 1024 (depth = max 2048)
	  {
	    register u_int32 tt = (u_int32)((u_int16)abs(phase)) * depth;
	    flangerSample = tt >> 16;
	    flangerPhase = tt;
	  }

	  // flangerdelay w/ added pre-delay
	  flgptr = bufptr + flangerSample - delay;
	  if (flgptr < &buffer[0])
	    flgptr += DELAY_BUF_SZ;

	  //USEX(SCI_VOL) = flangerValue;



	}

	/*Write ouput sample pairs*/
	WmaStereoCopy(audioSamp, BLOCKSIZE);

      } else {
#if 0
	//update variables from parametric struct
	depth = parametric_x.i.generic[4];
	speed = parametric_x.i.generic[5];
	delay = parametric_x.i.generic[6];
	decay = parametric_x.i.generic[7];
	mix = parametric_x.i.generic[8];
#endif
	//limit values (scale them??)
	if (depth > 2048)
	  depth = 2048;
	if (speed > 1024)
	  speed = 1024;
	if (speed == 0)
	  speed = 0;
	if (mix > 32000)
	  mix = 32000;
	if (decay > 32000)
	  decay = 32000;


#if 0
	// if modulation depth + pre-delay exeeds the delay buffer size 
	// make pre-delay shorter
	if (delay + depth > DELAY_BUF_SZ){
	  delay = DELAY_BUF_SZ - depth;
	}
#endif


      }

    }
}

