/*======================================================*/
/*==================== audio.c ==========================*/
/*=======================================================*/

//#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <vstypes.h>
#include <hardware.h>
//#include <math.h>


#define BLOCKSIZE 64 //Number of stereo samples processed in one loop round

#define SCI_MODE           0xC000
#define SCI_STATUS         0xC001
#define SCI_CLOCKF         0xC003
#define SCI_VOL         0xC00B
#define SCI_AICTRL0        0xC00C
#define SCI_AICTRL1        0xC00D
#define SCI_AICTRL2        0xC00E
#define SCI_AICTRL3        0xC00F
#define DECIM_FACTOR24K         6 // 24 kHz f256
#define DECIM_FACTOR48K         4 // 48 kHz f128
#define DECIM_FACTOR96K         2 // 96 kHz f64
#define DECIM_FACTOR192K        0 //192 kHz f32
#define DECIM_ENABLE            1
#define DECIM_CONTROL      0xc042
#define SCIST_AD_CLOCK          1
#define SCIST_APDOWN1           2 /* Analog internal */
#define SCIST_APDOWN2           3 /* Driver */
#define SCIMB_LINE             14
#define INT_ENABLE         0xC01A /* RW */
#define INT_EN_DAC              0
#define INT_EN_MODU             3
#define AUDIO_BUFFER_SZ      4096
#define STREAM_BUFFER_SZ     1024

auto void SetClockFromSpi(void);
auto void SetHardware(register s_int16 channels, register u_int16 rate);
auto s_int16 WmaStereoCopy(s_int16 *buf, s_int16 cnt);
auto void MyGetCPairs(register __i0 __y u_int16 *p, register __a0 s_int16 n);
extern s_int16 (*applAddr)(s_int16 register __i0 **d,
			   s_int16 register __a1 mode,
			   s_int16 register __a0 n);
void Disable(void);	/* NOT auto */
void Enable(void);	/* NOT auto */
extern u_int16 __y uartByteSpeed;
auto s_int16 StreamDiff(void);
auto s_int16 AudioBufFill(void);
extern s_int16 __y audio_buffer[];
volatile extern s_int16 __y * audio_wr_pointer;
volatile extern s_int16 __y * audio_rd_pointer;
extern u_int16 stream_buffer[];
volatile extern u_int16 __x * __y stream_wr_pointer;
volatile extern u_int16 __x * __y stream_rd_pointer;
u_int16 bp,hoopo;


/*
  Called before HALT, i.e. when there is nothing else to do.
  Either there is not enough data in stream buffer, or 
  the audio buffer is full.
 */
void UserHook(void) {
}

s_int16 __y buffer[8192]; //delay line

u_int16 decay; //(0 ... 32000U)
u_int16 delay; //(0 ... 8192)


void main(void) {
  s_int16 audioSamp[2*BLOCKSIZE];
  s_int16 __y lineInBuf[2*BLOCKSIZE];

  s_int16 __y *bufptr = buffer;

  s_int16 i = 0;

  bp=0;


  // adjust delay and decay
  delay = 3300;
  decay = 32000;


  //limit values
  if(delay > 8191)
    delay = 8191;

  if(decay > 55000)
    delay = 55000;

  //clear the delay line
  memsetY(buffer, 0, sizeof(buffer));

  Disable(); //Don't allow interrupts yet

    /* zero applAddr so we are not called again */
    applAddr = 0;

    /* must disable analog powerdown if it isn't already! */
    USEX(SCI_STATUS) &= ~((1<<SCIST_APDOWN1) | (1<<SCIST_APDOWN2));
    
    /*==========================================*/
    /*= Turn left analog input to line-in mode =*/
    /*==========================================*/
#if 0
    USEX(SCI_MODE) |= (1<<SCIMB_LINE); /*left=line-in, right=line-in*/
#else
    USEX(SCI_MODE) &= ~(1<<SCIMB_LINE); /*left=line-in, right=line-in*/
#endif

    /*================================================*/
    /*= Set uart speed for serial port communication =*/
    /*================================================*/

    //uartByteSpeed = 11520U;
    
    /*======================*/
    /*= Set DSP core clock =*/
    /*======================*/
    //USEX(SCI_CLOCKF) = 0x6000; /* 3.0x 12.288MHz */
    //USEX(SCI_CLOCKF) = 0x8000; /* 3.5x 12.288MHz */
    //USEX(SCI_CLOCKF) = 0xa000; /* 4.0x 12.288MHz */
    USEX(SCI_CLOCKF) = 0xc000; /* 4.5x 12.288MHz */
    //USEX(SCI_CLOCKF) = 0xe000; /* 5.0x 12.288MHz */
    SetClockFromSpi();

    USEX(SCI_AICTRL1) = 1024; /*autogain==0, 1x==1024, 0.5x==512 etc.*/
    USEX(SCI_AICTRL3) = 0x0004; /*Joint stereo (common gain). Linear pcm*/

    //adjust AD sample rate

    //USEX(DECIM_CONTROL) = DECIM_ENABLE | DECIM_FACTOR48K; /*AD sample rate is 48 kHz*/
    USEX(DECIM_CONTROL) = DECIM_ENABLE | DECIM_FACTOR24K; /*AD sample rate is 48 kHz*/
    
#define HALVE_AD_CLOCK 0 //Makes 12 kHz AD sample rate possible (24 kHz/2)
#if HALVE_AD_CLOCK
    USEX(SCI_STATUS) |= (1<<SCIST_AD_CLOCK); /*Halves AD clock 6MHz -> 3MHz */
#endif
    
    //SetHardware(2, 48000U); /*DA sample rate in Hz*/
    SetHardware(2, 24000U); /*DA sample rate in Hz*/

    /* 'Empty' stream buffer*/
    stream_rd_pointer = stream_wr_pointer = (u_int16 __x *)&stream_buffer[0];
    USEX(INT_ENABLE) |= (1<<INT_EN_MODU)/*AD*/ | (1<<INT_EN_DAC) | (1<<INT_EN_RX);

    USEX(SCI_VOL) = 0x0101;

    /* Set audio_buffer read and write pointers */
    audio_rd_pointer = audio_buffer;
    audio_wr_pointer = audio_buffer + 2*BLOCKSIZE;

    //Write zero to audio buffer. Prevent unwanted samples to be played.
    memsetY(audio_buffer, 0, AUDIO_BUFFER_SZ);

    Enable(); //allow interrupts (DAC, AD,...)

    while (1) { //Main loop
#if 0
      USEX(SCI_MODE) |= (1<<SCIMB_LINE); /*left=line-in, right=line-in*/
#endif
      if(StreamDiff() > BLOCKSIZE*2) {
	/*Read input samples*/
	MyGetCPairs((__y u_int16 *)lineInBuf, 2*BLOCKSIZE);

	{
	  int i;
	  __y s_int16 *lp = lineInBuf;
	  s_int16 *sp = audioSamp;
	  
	  for(i=0;i<BLOCKSIZE;i++) {


	    //add input sample to buffer with old value
	    *bufptr = ((s_int32)*bufptr * decay + (s_int32)lp[0] * 32000U) >> 16;
	    lp += 2;

	    //buffer to output (mono out...)
	    *sp++ = *bufptr;
	    *sp++ = *bufptr;

	    //wrap buffer and adjust delay time
	    if (bufptr  >= &buffer[0] + delay) {
	      bufptr = buffer;
	    } else {
	      bufptr++;
	    }

	  }

	}

	/*Write ouput sample pairs*/
	WmaStereoCopy(audioSamp, BLOCKSIZE);
      }

    }
}

